@@ -1,13 +1,39 @@
 import string
 
 from django.db import models
+from django.db.models import query
 from django.contrib.contenttypes.models import ContentType
 from django.contrib.contenttypes import generic
 from django.core.exceptions import ObjectDoesNotExist
 from django.utils.translation import ugettext_lazy as _
+from django.core.cache import cache
+
+JOURNAL_METADATA_CACHE_TIMEOUT = 3600
+
+class CachedQuerySet(query.QuerySet):
+    def get_cached(self, **kwargs):
+        key = tuple(sorted(kwargs.iteritems()))
+        hash1 = key.__hash__()
+        hash2 = ('x', key).__hash__()
+        key = 'dj-cache-%s-%s-%s' % (self.model.__name__, hash1, hash2)
+        instance = cache.get(key)
+        if instance is not None:
+            return instance
+        instance, created = self.get_or_create(**kwargs)
+        if JOURNAL_METADATA_CACHE_TIMEOUT:
+            cache.set(key, instance, JOURNAL_METADATA_CACHE_TIMEOUT)
+        return instance
 
 
-class TagManager(models.Manager):
+class CachedManager(models.Manager):
+    def get_query_set(self):
+        return CachedQuerySet(self.model, using=self._db)
+
+    def get_cached(self, **kwargs):
+        return self.get_query_set().get_cached(**kwargs)
+
+
+class TagManager(CachedManager):
     def get_by_natural_key(self, name):
         return self.get(name=name)
 
@@ -32,7 +58,7 @@ class Tag(models.Model):
         verbose_name = _('Journal tag')
 
 
-class TemplateManager(models.Manager):
+class TemplateManager(CachedManager):
     def get_by_natural_key(self, content):
         return self.get(content=content)
 
@@ -57,6 +83,14 @@ class Template(models.Model):
 
 
 class JournalManager(models.Manager):
+    def get_query_set(self):
+        return super(JournalManager, self).get_query_set() \
+               .prefetch_related('objectdata_set__content_type',
+                       'stringdata_set', 'objectdata_set__tag',
+                       'stringdata_set__tag', 'objectdata_set__content_object',
+                       'tag', 'template') \
+               .select_related('tag', 'template')
+
     def for_object(self, obj, tag=None):
         '''Return Journal records linked to this object.'''
         content_type = ContentType.objects.get_for_model(obj)
@@ -91,7 +125,7 @@ class JournalManager(models.Manager):
         '''
         if not isinstance(tag, Tag):
             try:
-                tag = Tag.objects.get(name=tag)
+                tag = Tag.objects.get_cached(name=tag)
             except Tag.DoesNotExist:
                 return self.none()
         # always remember: multiple join (OR in WHERE) produces duplicate
@@ -128,10 +162,10 @@ class Journal(models.Model):
     def message_context(self):
         ctx = {}
         for data in self.objectdata_set.all():
-            try:
+            if data.content_object is not None:
                 ctx[data.tag.name] = data.content_object
-            except ObjectDoesNotExist:
-                ctx[data.tag.name] = '<deleted {content_type} {object_id}>'.format(
+            else:
+                ctx[data.tag.name] = u'<deleted {content_type} {object_id}>'.format(
                         content_type=data.content_type, object_id=data.object_id)
         for data in self.stringdata_set.all():
             ctx[data.tag.name] = data.content
